@@ -15,11 +15,10 @@
 
 //! Execution engine for EndBASIC programs.
 
-use crate::ast::{ArgSep, Expr, Statement, Value, VarRef, VarType};
-use crate::eval::{self, CallableMetadata, Function, Vars};
+use crate::ast::{Expr, Statement, Value, VarRef, VarType};
+use crate::eval;
 use crate::parser::{self, Parser};
-use async_trait::async_trait;
-use std::collections::HashMap;
+use crate::syms::{CallError, CallableMetadata, Command, Function, Symbol, Symbols};
 use std::future::Future;
 use std::io;
 use std::pin::Pin;
@@ -43,10 +42,26 @@ pub enum Error {
     /// Syntax error.
     #[error("{0}")]
     SyntaxError(String),
+}
 
-    /// Execution of a builtin command failed due to a user call error.
-    #[error("{0}")]
-    UsageError(String),
+impl Error {
+    /// Annotates a call evaluation error with the command's metadata.
+    // TODO(jmmv): This is a hack to support the transition to a better Command abstraction within
+    // Symbols and exists to minimize the amount of impacted tests.  Should be removed and/or
+    // somehow unified with the equivalent function in eval::Error.
+    pub(crate) fn from_call_error(md: &CallableMetadata, e: CallError) -> Self {
+        match e {
+            CallError::ArgumentError(e) => Self::SyntaxError(e),
+            CallError::EvalError(e) => Self::EvalError(e),
+            CallError::InternalError(e) => Self::SyntaxError(e),
+            CallError::IoError(e) => Self::IoError(e),
+            CallError::SyntaxError => Self::SyntaxError(format!(
+                "Syntax error in call to {}: expected {}",
+                md.name(),
+                md.syntax()
+            )),
+        }
+    }
 }
 
 /// Result for execution return values.
@@ -57,40 +72,6 @@ fn new_syntax_error<T, S: Into<String>>(message: S) -> Result<T> {
     Err(Error::SyntaxError(message.into()))
 }
 
-/// Instantiates a new `Err(Error::UsageError(...))` from a message.  Syntactic sugar.
-pub fn new_usage_error<T, S: Into<String>>(message: S) -> Result<T> {
-    Err(Error::UsageError(message.into()))
-}
-
-/// A trait to define a command that is executed by a `Machine`.
-///
-/// The commands themselves are immutable but they can reference mutable state.  Given that
-/// EndBASIC is not threaded, it is sufficient for those references to be behind a `RefCell`
-/// and/or an `Rc`.
-///
-/// Idiomatically, these objects need to provide a `new()` method that returns an `Rc<Callable>`, as
-/// that's the type used throughout the execution engine.
-#[async_trait(?Send)]
-pub trait Command {
-    /// Returns the metadata for this command.
-    ///
-    /// The return value takes the form of a reference to force the callable to store the metadata
-    /// as a struct field so that calls to this function are guaranteed to be cheap.
-    fn metadata(&self) -> &CallableMetadata;
-
-    /// Executes the command.
-    ///
-    /// `args` contains the arguments as provided in the invocation of the command.  Each entry in
-    /// this array contains an optional expression (to support things like `PRINT a, , b`) and the
-    /// separator that was used between that argument and the next.  The last entry in `args` always
-    /// has `ArgSep::End` as the separator.
-    ///
-    /// `machine` provides mutable access to the current state of the machine invoking the command.
-    ///
-    /// Commands cannot return any value except for errors.
-    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()>;
-}
-
 /// Describes how the machine stopped execution while it was running a script via `exec()`.
 #[derive(Clone, Debug, PartialEq)]
 #[must_use]
@@ -115,36 +96,24 @@ impl StopReason {
 /// Executes an EndBASIC program and tracks its state.
 #[derive(Default)]
 pub struct Machine {
-    commands: HashMap<&'static str, Rc<dyn Command>>,
-    functions: HashMap<&'static str, Rc<dyn Function>>,
-    vars: Vars,
+    symbols: Symbols,
     stop_reason: Option<StopReason>,
 }
 
 impl Machine {
     /// Registers the given builtin command, which must not yet be registered.
     pub fn add_command(&mut self, command: Rc<dyn Command>) {
-        let metadata = command.metadata();
-        assert!(
-            self.commands.get(&metadata.name()).is_none(),
-            "Command with the same name already registered"
-        );
-        self.commands.insert(metadata.name(), command);
+        self.symbols.add_command(command)
     }
 
     /// Registers the given builtin function, which must not yet be registered.
     pub fn add_function(&mut self, function: Rc<dyn Function>) {
-        let metadata = function.metadata();
-        assert!(
-            self.functions.get(&metadata.name()).is_none(),
-            "Function with the same name already registered"
-        );
-        self.functions.insert(metadata.name(), function);
+        self.symbols.add_function(function)
     }
 
     /// Resets the state of the machine by clearing all variable.
     pub fn clear(&mut self) {
-        self.vars.clear()
+        self.symbols.clear()
     }
 
     /// Tells the machine to stop execution at the next statement boundary.
@@ -155,30 +124,20 @@ impl Machine {
         self.stop_reason = Some(StopReason::Exited(code));
     }
 
-    /// Obtains immutable access to the builtin commands provided by this machine.
-    pub fn get_commands(&self) -> &HashMap<&'static str, Rc<dyn Command>> {
-        &self.commands
+    /// Obtains immutable access to the state of the symbols.
+    pub fn get_symbols(&self) -> &Symbols {
+        &self.symbols
     }
 
-    /// Obtains immutable access to the builtin functions provided by this machine.
-    pub fn get_functions(&self) -> &HashMap<&'static str, Rc<dyn Function>> {
-        &self.functions
-    }
-
-    /// Obtains immutable access to the state of the variables.
-    pub fn get_vars(&self) -> &Vars {
-        &self.vars
-    }
-
-    /// Obtains mutable access to the state of the variables.
-    pub fn get_mut_vars(&mut self) -> &mut Vars {
-        &mut self.vars
+    /// Obtains mutable access to the state of the symbols.
+    pub fn get_mut_symbols(&mut self) -> &mut Symbols {
+        &mut self.symbols
     }
 
     /// Retrieves the variable `name` as a boolean.  Fails if it is some other type or if it's not
     /// defined.
     pub fn get_var_as_bool(&self, name: &str) -> Result<bool> {
-        match self.vars.get(&VarRef::new(name, VarType::Boolean))? {
+        match self.symbols.get_var(&VarRef::new(name, VarType::Boolean))? {
             Value::Boolean(b) => Ok(*b),
             _ => panic!("Invalid type check in get()"),
         }
@@ -187,7 +146,7 @@ impl Machine {
     /// Retrieves the variable `name` as an integer.  Fails if it is some other type or if it's not
     /// defined.
     pub fn get_var_as_int(&self, name: &str) -> Result<i32> {
-        match self.vars.get(&VarRef::new(name, VarType::Integer))? {
+        match self.symbols.get_var(&VarRef::new(name, VarType::Integer))? {
             Value::Integer(i) => Ok(*i),
             _ => panic!("Invalid type check in get()"),
         }
@@ -196,7 +155,7 @@ impl Machine {
     /// Retrieves the variable `name` as a string.  Fails if it is some other type or if it's not
     /// defined.
     pub fn get_var_as_string(&self, name: &str) -> Result<&str> {
-        match self.vars.get(&VarRef::new(name, VarType::Text))? {
+        match self.symbols.get_var(&VarRef::new(name, VarType::Text))? {
             Value::Text(s) => Ok(s),
             _ => panic!("Invalid type check in get()"),
         }
@@ -204,15 +163,56 @@ impl Machine {
 
     /// Assigns the value of `expr` to the variable `vref`.
     fn assign(&mut self, vref: &VarRef, expr: &Expr) -> Result<()> {
-        let value = expr.eval(&self.vars, &self.functions)?;
-        self.vars.set(&vref, value)?;
+        let value = expr.eval(&mut self.symbols)?;
+        self.symbols.set_var(&vref, value)?;
+        Ok(())
+    }
+
+    /// Assigns the value of `expr` to the array `vref` in the position `subscripts`.
+    fn assign_array(&mut self, vref: &VarRef, subscripts: &[Expr], expr: &Expr) -> Result<()> {
+        let mut ds = Vec::with_capacity(subscripts.len());
+        for ss_expr in subscripts {
+            match ss_expr.eval(&mut self.symbols)? {
+                Value::Integer(i) => ds.push(i),
+                v => return new_syntax_error(format!("Subscript {:?} must be an integer", v)),
+            }
+        }
+
+        let value = expr.eval(&mut self.symbols)?;
+
+        match self.symbols.get_mut(vref)? {
+            Some(Symbol::Array(array)) => {
+                array.assign(&ds, value)?;
+                Ok(())
+            }
+            Some(_) => new_syntax_error(format!("Cannot index non-array {}", vref.name())),
+            None => new_syntax_error(format!("Cannot index undefined array {}", vref.name())),
+        }
+    }
+
+    /// Defines a new array `name` of type `subtype` with `dimensions`.  The array must not yet
+    /// exist, and the name may not overlap function or variable names.
+    pub fn dim_array(&mut self, name: &str, subtype: &VarType, dimensions: &[Expr]) -> Result<()> {
+        let mut ds = Vec::with_capacity(dimensions.len());
+        for dim_expr in dimensions {
+            match dim_expr.eval(&mut self.symbols)? {
+                Value::Integer(i) => {
+                    if i <= 0 {
+                        return new_syntax_error("Dimensions in DIM array must be positive");
+                    }
+                    ds.push(i as usize);
+                }
+                _ => return new_syntax_error("Dimensions in DIM array must be integers"),
+            }
+        }
+        self.symbols.dim_array(name, *subtype, ds)?;
         Ok(())
     }
 
     /// Executes an `IF` statement.
     async fn do_if(&mut self, branches: &[(Expr, Vec<Statement>)]) -> Result<()> {
         for (expr, stmts) in branches {
-            match expr.eval(&self.vars, &self.functions)? {
+            match expr.eval(&mut self.symbols)? {
                 Value::Boolean(true) => {
                     for s in stmts {
                         self.exec_one(s).await?;
@@ -238,14 +238,14 @@ impl Machine {
         debug_assert!(
             iterator.ref_type() == VarType::Auto || iterator.ref_type() == VarType::Integer
         );
-        let start_value = start.eval(&self.vars, &self.functions)?;
+        let start_value = start.eval(&mut self.symbols)?;
         match start_value {
-            Value::Integer(_) => self.vars.set(iterator, start_value)?,
+            Value::Integer(_) => self.symbols.set_var(iterator, start_value)?,
             _ => return new_syntax_error("FOR supports integer iteration only"),
         }
 
         loop {
-            match end.eval(&self.vars, &self.functions)? {
+            match end.eval(&mut self.symbols)? {
                 Value::Boolean(false) => {
                     break;
                 }
@@ -265,7 +265,7 @@ impl Machine {
     /// Executes a `WHILE` loop.
     async fn do_while(&mut self, condition: &Expr, body: &[Statement]) -> Result<()> {
         loop {
-            match condition.eval(&self.vars, &self.functions)? {
+            match condition.eval(&mut self.symbols)? {
                 Value::Boolean(true) => {
                     for s in body {
                         self.exec_one(s).await?;
@@ -285,13 +285,23 @@ impl Machine {
         }
 
         match stmt {
+            Statement::ArrayAssignment(vref, subscripts, value) => {
+                self.assign_array(vref, subscripts, value)?
+            }
             Statement::Assignment(vref, expr) => self.assign(vref, expr)?,
             Statement::BuiltinCall(name, args) => {
-                let cmd = match self.commands.get(name.as_str()) {
-                    Some(cmd) => cmd.clone(),
+                let cmd = match self.symbols.get(&VarRef::new(name, VarType::Auto))? {
+                    Some(Symbol::Command(cmd)) => cmd.clone(),
+                    Some(_) => return new_syntax_error(format!("{} is not a command", name)),
                     None => return new_syntax_error(format!("Unknown builtin {}", name)),
                 };
-                cmd.exec(&args, self).await?
+                cmd.exec(&args, self)
+                    .await
+                    .map_err(|e| Error::from_call_error(cmd.metadata(), e))?;
+            }
+            Statement::Dim(varname, vartype) => self.symbols.dim(varname, *vartype)?,
+            Statement::DimArray(varname, dimensions, subtype) => {
+                self.dim_array(varname, subtype, dimensions)?
             }
             Statement::If(branches) => {
                 // Change this to using FutureExt::boxed_local if we ever depend on the futures or
@@ -334,145 +344,10 @@ impl Machine {
     }
 }
 
-#[cfg(test)]
-pub(crate) mod testutils {
-    use super::*;
-    use crate::eval::CallableMetadataBuilder;
-    use std::cell::RefCell;
-
-    /// Simplified version of `EXIT` to test the machine's `exit()` method.
-    pub struct ExitCommand {
-        metadata: CallableMetadata,
-    }
-
-    impl ExitCommand {
-        /// Creates a new command that terminates execution once called.
-        pub fn new() -> Rc<Self> {
-            Rc::from(Self {
-                metadata: CallableMetadataBuilder::new("EXIT", VarType::Void).test_build(),
-            })
-        }
-    }
-
-    #[async_trait(?Send)]
-    impl Command for ExitCommand {
-        fn metadata(&self) -> &CallableMetadata {
-            &self.metadata
-        }
-
-        async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()> {
-            let arg = match args {
-                [(Some(expr), ArgSep::End)] => {
-                    match expr.eval(machine.get_vars(), machine.get_functions())? {
-                        Value::Integer(n) => {
-                            assert!((0..128).contains(&n), "Exit code out of range");
-                            n as u8
-                        }
-                        _ => panic!("Exit code must be a positive integer"),
-                    }
-                }
-                _ => panic!("EXIT takes one argument"),
-            };
-            machine.exit(arg);
-            Ok(())
-        }
-    }
-
-    /// Simplified version of `INPUT` to feed input values based on some golden `data`.
-    ///
-    /// Every time this command is invoked, it yields the next value from the `data` iterator and
-    /// assigns it to the variable provided as its only argument.
-    pub(crate) struct InCommand {
-        metadata: CallableMetadata,
-        data: Box<RefCell<dyn Iterator<Item = &'static &'static str>>>,
-    }
-
-    impl InCommand {
-        /// Creates a new command with the golden `data`.
-        pub(crate) fn new(
-            data: Box<RefCell<dyn Iterator<Item = &'static &'static str>>>,
-        ) -> Rc<Self> {
-            Rc::from(Self {
-                metadata: CallableMetadataBuilder::new("IN", VarType::Void).test_build(),
-                data,
-            })
-        }
-    }
-
-    #[async_trait(?Send)]
-    impl Command for InCommand {
-        fn metadata(&self) -> &CallableMetadata {
-            &self.metadata
-        }
-
-        async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()> {
-            if args.len() != 1 {
-                return new_usage_error("IN only takes one argument");
-            }
-            if args[0].1 != ArgSep::End {
-                return new_usage_error("Invalid separator");
-            }
-            let vref = match &args[0].0 {
-                Some(Expr::Symbol(vref)) => vref,
-                _ => return new_usage_error("IN requires a variable reference"),
-            };
-
-            let mut data = self.data.borrow_mut();
-            let raw_value = data.next().unwrap().to_owned();
-            let value = Value::parse_as(vref.ref_type(), raw_value)?;
-            machine.get_mut_vars().set(vref, value)?;
-            Ok(())
-        }
-    }
-
-    /// Simplified version of `PRINT` that captures all calls to it into `data`.
-    ///
-    /// This command only accepts arguments separated by the `;` short separator and concatenates
-    /// them with a single space.
-    pub(crate) struct OutCommand {
-        metadata: CallableMetadata,
-        data: Rc<RefCell<Vec<String>>>,
-    }
-
-    impl OutCommand {
-        /// Creates a new command that captures all calls into `data`.
-        pub(crate) fn new(data: Rc<RefCell<Vec<String>>>) -> Rc<Self> {
-            Rc::from(Self {
-                metadata: CallableMetadataBuilder::new("OUT", VarType::Void).test_build(),
-                data,
-            })
-        }
-    }
-
-    #[async_trait(?Send)]
-    impl Command for OutCommand {
-        fn metadata(&self) -> &CallableMetadata {
-            &self.metadata
-        }
-
-        async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()> {
-            let mut text = String::new();
-            for arg in args.iter() {
-                if let Some(expr) = arg.0.as_ref() {
-                    text += &expr.eval(machine.get_vars(), machine.get_functions())?.to_string();
-                }
-                match arg.1 {
-                    ArgSep::End => break,
-                    ArgSep::Short => text += " ",
-                    ArgSep::Long => return new_usage_error("Cannot use the ',' separator"),
-                }
-            }
-            self.data.borrow_mut().push(text);
-            Ok(())
-        }
-    }
-}
-
 #[cfg(test)]
 mod tests {
-    use super::testutils::*;
     use super::*;
-    use crate::eval::testutils::*;
+    use crate::testutils::*;
     use futures_lite::future::block_on;
     use std::cell::RefCell;
     use std::rc::Rc;
@@ -604,6 +479,35 @@ mod tests {
         do_error_test(input, &[], &[], expected_err);
     }
 
+    #[test]
+    fn test_array_assignment_ok() {
+        do_ok_test("DIM a(3)\na(1) = 5 + 1\nOUT a(0); a(1); a(2)", &[], &["0 6 0"]);
+        do_ok_test("DIM a(3) AS STRING\na$(1) = \"x\"\nOUT a(0); a(1); a$(2)", &[], &[" x "]);
+        do_ok_test(
+            "DIM a(3, 8, 2) AS BOOLEAN\na(3 - 3, 2 * 2, 1) = TRUE\nOUT a(0, 4, 1)",
+            &[],
+            &["TRUE"],
+        );
+    }
+
+    #[test]
+    fn test_array_assignment_ok_case_insensitive() {
+        do_ok_test("DIM a(3)\nA(1) = 5\na(2) = 1\nOUT A(0); a(1); A(2)", &[], &["0 5 1"]);
+    }
+
+    #[test]
+    fn test_array_assignment_errors() {
+        do_simple_error_test("a() = 3\n", "Cannot index undefined array a");
+        do_simple_error_test("a = 3\na(0) = 3\n", "Cannot index non-array a");
+        do_simple_error_test("DIM a(2)\na() = 3\n", "Cannot index array with 0 subscripts; need 1");
+        do_simple_error_test("DIM a(1)\na(-1) = 3\n", "Subscript -1 cannot be negative");
+        do_simple_error_test(
+            "DIM a(1)\na(1, 3.0) = 3\n",
+            "Subscript Double(3.0) must be an integer",
+        );
+        do_simple_error_test("DIM a(2)\na$(1) = 3", "Incompatible types in a$ reference");
+    }
+
     #[test]
     fn test_assignment_ok_types() {
         do_ok_test("a = TRUE\nOUT a; a?", &[], &["TRUE TRUE"]);
@@ -631,6 +535,35 @@ mod tests {
         do_simple_error_test("a? = 3", "Incompatible types in a? assignment");
     }
 
+    #[test]
+    fn test_dim_ok() {
+        do_ok_test("DIM foo\nDIM bar AS BOOLEAN\nOUT foo%; bar?", &[], &["0 FALSE"]);
+    }
+
+    #[test]
+    fn test_dim_errors() {
+        do_simple_error_test("DIM i\nDIM i", "Cannot DIM already-defined symbol i");
+        do_simple_error_test("DIM i\nDIM I", "Cannot DIM already-defined symbol I");
+        do_simple_error_test("i = 0\nDIM i", "Cannot DIM already-defined symbol i");
+    }
+
+    #[test]
+    fn test_dim_array_ok() {
+        do_ok_test(
+            "DIM foo(3, 4)\nDIM bar(1) AS BOOLEAN\nOUT foo%(2, 2); bar?(0)",
+            &[],
+            &["0 FALSE"],
+        );
+    }
+
+    #[test]
+    fn test_dim_array_errors() {
+        do_simple_error_test("DIM i()", "Arrays require at least one dimension");
+        do_simple_error_test("DIM i(FALSE)", "Dimensions in DIM array must be integers");
+        do_simple_error_test("DIM i(-3)", "Dimensions in DIM array must be positive");
+        do_simple_error_test("DIM i\nDIM i(3)", "Cannot DIM already-defined symbol i");
+    }
+
     #[test]
     fn test_exit_simple() {
         let captured_out = Rc::from(RefCell::from(vec![]));
@@ -854,7 +787,8 @@ mod tests {
 
     #[test]
     fn test_function_call_errors() {
-        do_simple_error_test("OUT SUM?()", "Incompatible type annotation for function call");
+        do_simple_error_test("OUT OUT()", "OUT is not an array or a function");
+        do_simple_error_test("OUT SUM?()", "Incompatible types in SUM? reference");
     }
 
     #[test]
@@ -864,21 +798,21 @@ mod tests {
             WHILE n > 0
                 OUT "n is"; n
                 n = n - 1
-            END WHILE
+            WEND
         "#;
         do_ok_test(code, &["0"], &[]);
         do_ok_test(code, &["3"], &["n is 3", "n is 2", "n is 1"]);
 
-        do_ok_test("WHILE FALSE\nOUT 1\nEND WHILE", &[], &[]);
+        do_ok_test("WHILE FALSE\nOUT 1\nWEND", &[], &[]);
     }
 
     #[test]
     fn test_while_errors() {
-        do_simple_error_test("WHILE\nEND WHILE", "No expression in WHILE statement");
-        do_simple_error_test("WHILE\nEND IF", "WHILE without END WHILE");
+        do_simple_error_test("WHILE\nWEND", "No expression in WHILE statement");
+        do_simple_error_test("WHILE\nEND IF", "WHILE without WEND");
 
-        do_simple_error_test("WHILE 2\n", "WHILE without END WHILE");
-        do_simple_error_test("WHILE 2\nEND WHILE", "WHILE requires a boolean condition");
+        do_simple_error_test("WHILE 2\n", "WHILE without WEND");
+        do_simple_error_test("WHILE 2\nWEND", "WHILE requires a boolean condition");
     }
 
     #[test]
