@@ -70,7 +70,15 @@ pub enum Token {
     Step,
     Then,
     To,
+    Wend,
     While,
+
+    Dim,
+    As,
+    BooleanName,
+    DoubleName,
+    IntegerName,
+    TextName,
 }
 
 /// Extra operations to test properties of a `char` based on the language semantics.
@@ -261,21 +269,28 @@ impl<'a> Lexer<'a> {
         }
         match s.to_uppercase().as_str() {
             "AND" => Ok(Token::And),
+            "AS" => Ok(Token::As),
+            "BOOLEAN" => Ok(Token::BooleanName),
+            "DIM" => Ok(Token::Dim),
+            "DOUBLE" => Ok(Token::DoubleName),
             "ELSE" => Ok(Token::Else),
             "ELSEIF" => Ok(Token::Elseif),
             "END" => Ok(Token::End),
             "FALSE" => Ok(Token::Boolean(false)),
             "FOR" => Ok(Token::For),
             "IF" => Ok(Token::If),
+            "INTEGER" => Ok(Token::IntegerName),
             "MOD" => Ok(Token::Modulo),
             "NEXT" => Ok(Token::Next),
             "NOT" => Ok(Token::Not),
             "OR" => Ok(Token::Or),
             "REM" => self.consume_rest_of_line(),
             "STEP" => Ok(Token::Step),
+            "STRING" => Ok(Token::TextName),
             "THEN" => Ok(Token::Then),
             "TO" => Ok(Token::To),
             "TRUE" => Ok(Token::Boolean(true)),
+            "WEND" => Ok(Token::Wend),
             "WHILE" => Ok(Token::While),
             "XOR" => Ok(Token::Xor),
             _ => Ok(Token::Symbol(VarRef::new(s, vtype))),
@@ -583,6 +598,21 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_dim() {
+        do_ok_test("DIM AS", &[Token::Dim, Token::As]);
+        do_ok_test(
+            "BOOLEAN DOUBLE INTEGER STRING",
+            &[Token::BooleanName, Token::DoubleName, Token::IntegerName, Token::TextName],
+        );
+
+        do_ok_test("dim as", &[Token::Dim, Token::As]);
+        do_ok_test(
+            "boolean double integer string",
+            &[Token::BooleanName, Token::DoubleName, Token::IntegerName, Token::TextName],
+        );
+    }
+
     #[test]
     fn test_if() {
         do_ok_test(
@@ -605,9 +635,9 @@ mod tests {
 
     #[test]
     fn test_while() {
-        do_ok_test("WHILE END WHILE", &[Token::While, Token::End, Token::While]);
+        do_ok_test("WHILE WEND", &[Token::While, Token::Wend]);
 
-        do_ok_test("while end while", &[Token::While, Token::End, Token::While]);
+        do_ok_test("while wend", &[Token::While, Token::Wend]);
     }
 
     /// Syntactic sugar to instantiate a test that verifies the parsing of an operator.
