@@ -468,7 +468,7 @@ impl CallableMetadataBuilder {
 ///
 /// The callable is expected to hold onto an instance of this object within its struct to make
 /// queries fast.
-#[derive(Debug)]
+#[derive(Clone, Debug)]
 pub struct CallableMetadata {
     name: &'static str,
     return_type: VarType,
@@ -493,7 +493,8 @@ impl CallableMetadata {
         self.syntax
     }
 
-    /// Gets the callable's category.
+    /// Gets the callable's category as a collection of lines.  The first line is the title of the
+    /// category, and any extra lines are additional information for it.
     pub fn category(&self) -> &'static str {
         self.category
     }
@@ -521,9 +522,13 @@ pub trait Function {
 
     /// Executes the function.
     ///
-    /// `args` contains the evaluated arguments as provided in the invocation of the function.
+    /// `args` contains the unevaluated arguments as provided in the invocation of the function.
+    /// These are needed to support functions that need unevaluated symbol references (such as
+    /// `LBOUND(array)`).  Because most functions don't support this kind of input, they should use
+    /// `eval::eval_all` to process all arguments.
+    ///
     /// `symbols` provides mutable access to the current state of the machine's symbols.
-    fn exec(&self, args: Vec<Value>, symbols: &mut Symbols) -> FunctionResult;
+    fn exec(&self, args: &[Expr], symbols: &mut Symbols) -> FunctionResult;
 }
 
 /// A trait to define a command that is executed by a `Machine`.
